from __future__ import absolute_import
from collections import OrderedDict
import inspect

__all__ = ["merge_numpy_docs"]


def parse_numpy_doc(doc):
    """ Extract the text from the various sections of a numpy-formatted docstring.

        Parameters
        ----------
        doc: Union[str, None]

        Returns
        -------
        OrderedDict[str, Union[None,str]]
            The extracted numpy-styled docstring sections."""

    doc_sections = OrderedDict([("Short Summary", None),
                                ("Deprecation Warning", None),
                                ("Parameters", None),
                                ("Attributes", None),
                                ("Extended Summary", None),
                                ("Parameters", None),
                                ("Returns", None),
                                ("Yields", None),
                                ("Other Parameters", None),
                                ("Raises", None),
                                ("See Also", None),
                                ("Notes", None),
                                ("References", None),
                                ("Examples", None)])

    if not doc:
        return doc_sections

    doc = inspect.cleandoc(doc)
    lines = iter(doc.splitlines())

    key = "Short Summary"
    body = []
    while True:
        try:
            line = next(lines)
            if line in doc_sections:
                doc_sections[key] = "\n".join(body).rstrip() if body else None
                body = []
                key = line
                next(lines)  # skip section delimiter
            else:
                body.append(line)
        except StopIteration:
            doc_sections[key] = "\n".join(body)
            break

    return doc_sections


def merge_section(key, prnt_sec, child_sec):
    """ Synthesize a output numpy docstring section.

        Parameters
        ----------
        key: str
            The numpy-section being merged.
        prnt_sec: Optional[str]
            The docstring section from the parent's attribute.
        child_sec: Optional[str]
            The docstring section from the child's attribute.
        Returns
        -------
        Optional[str]
            The output docstring section."""
    if prnt_sec is None and child_sec is None:
        return None

    if key == "Short Summary":
        header = ''
    else:
        header = "\n".join((key, "".join("-" for i in range(len(key))), ""))

    if child_sec is None:
        body = prnt_sec
    else:
        body = child_sec

    return header + body


def merge_all_sections(prnt_sctns, child_sctns):
    """ Merge the doc-sections of the parent's and child's attribute into a single docstring.

        Parameters
        ----------
        prnt_sctns: OrderedDict[str, Union[None,str]]
        child_sctns: OrderedDict[str, Union[None,str]]

        Returns
        -------
        str
            Output docstring of the merged docstrings."""
    doc = []

    prnt_only_raises = prnt_sctns["Raises"] and not (prnt_sctns["Returns"] or prnt_sctns["Yields"])
    if prnt_only_raises and (child_sctns["Returns"] or child_sctns["Yields"]):
        prnt_sctns["Raises"] = None

    for key in prnt_sctns:
        sect = merge_section(key, prnt_sctns[key], child_sctns[key])
        if sect is not None:
            doc.append(sect)
    return "\n\n".join(doc) if doc else None


def merge_numpy_docs(prnt_doc=None, child_doc=None):
    """ Merge two numpy-style docstrings into a single docstring.

        Given the numpy-style docstrings from a parent and child's attributes, merge the docstring
        sections such that the child's section is used, wherever present, otherwise the parent's
        section is used.

        Parameters
        ----------
        prnt_doc: Optional[str]
            The docstring from the parent.
        child_doc: Optional[str]
            The docstring from the child.

        Returns
        -------
        Optional[str]
            The merged docstring.
        """
    return merge_all_sections(parse_numpy_doc(prnt_doc), parse_numpy_doc(child_doc))



from __future__ import absolute_import
from .doc_parse_tools import merge_numpy_docs

""" Docstring inheritance-style implementations.

    To implement your own inheritance file, simply write a function that fits the template:

            def your_style(prnt_doc, child_doc):
                ''' Merges parent and child docstrings

                    Parameters
                    ----------
                    prnt_cls_doc: Optional[str]
                    child_doc: Optional[str]

                    Returns
                    ------
                    Optional[str]
                        The merged docstring that will be utilized.'''
                return final_docstring

    and log this using `custom_inherit.add_style(your_style)`. To permanently save your function,
    define your function within custom_inherit/style_store.py, and log it in custom_inherit.style_store.__all__.
    Your style will then be available as 'your_style' (i.e. whatever you named the function).


    The built-in styles are:
        - parent:   Wherever the docstring for a child-class' attribute (or for the class itself) is
                    `None`, inherit the corresponding docstring from the parent.

                    *NOTE* As of Python 3.5, this is the default behavior of the built-in function
                    inspect.getdoc, and thus this style is deprecated Python 3.(>=5).

        - numpy:    The numpy-styled docstrings from the parent and child are merged gracefully
                    with nice formatting.

                    Specifically, any docstring section that appears in the parent's docstring that
                    is not present in the child's is inherited. Otherwise, the child's docstring
                    section is utilized. An exception to this is if the parent docstring contains a
                    "Raises" section, but the child's attribute's docstring contains a "Returns" or
                    "Yields" section instead. In this instance, the "Raises" section will not appear in the
                    inherited docstring.

                    Example:
                        parent-attribute's docstring:

                            ''' Parent's line

                                Parameters
                                ----------
                                x: int
                                    blah-x
                                y: Union[None, int]
                                    blah-y

                                Raises
                                -------
                                NotImplemented Error'''

                        docstring for corresponding attribute of child:

                            ''' Child's line

                                Returns
                                -------
                                int

                                Notes
                                -----
                                notes blah blah'''

                        docstring that is ultimately inherited by child's attribute:

                            ''' Child's line

                                Parameters
                                ----------
                                x: int
                                    blah-x
                                y: Union[None, int]
                                    blah-y

                                Returns
                                -------
                                int

                                Notes
                                -----
                                notes blah blah'''
"""

# All built-in styles must be logged in the __all__ field.
__all__ = ["parent", "numpy"]


def parent(prnt_doc, child_doc):
    return child_doc if child_doc is not None else prnt_doc


def numpy(prnt_doc, child_doc):
    return merge_numpy_docs(prnt_doc, child_doc)



from __future__ import absolute_import
from abc import ABCMeta
from .metaclass_base import DocInheritorBase
from .style_store import *
from .decorator_base import DocInheritDecorator
from .style_store import __all__ as all_styles

try:
    basestring
except NameError:
    basestring = str  # Python 2 -> 3 alias


__all__ = ["DocInheritMeta", "doc_inherit", "store", "add_style", "remove_style"]
__version__ = "2.0.2"


def _check_style_function(style_func):
    out = style_func("", "")
    if not isinstance(out, basestring) and out is not None:
        raise TypeError
    return None


class _Store(object):
    """ A dictionary-like object that stores the styles available for the doc-inheritance metaclass and decorator,
       respectively.

       Only callable objects with the signature: f(Optional[str], Optional[str]) -> Optional[str]
       can be stored. If f is a valid callable, then _Store()[f] -> f."""

    def __init__(self, *args, **kwargs):
        self._store = dict()
        self.update(*args, **kwargs)

    def __repr__(self):
        return repr(self._store)

    def __str__(self):
        out_str = "The available stored styles are: "
        styles = "\n".join("\t- " + style for style in sorted(self.keys()))
        return "\n".join((out_str, styles))

    def __setitem__(self, style_name, style_func):
        """ Make available a new function for merging a 'parent' and 'child' docstring.

            Parameters
            ----------
            style_name : Any
                The identifier of the style being logged
            style_func: Callable[[Optional[str], Optional[str]], Optional[str]]
                The style function that merges two docstrings into a single docstring."""
        try:
            _check_style_function(style_func)
        except TypeError:
            raise TypeError("The style store only stores callables of the form: "
                            "\n\tstyle_func(Optional[str], Optional[str]) -> Optional[str]")
        self._store[style_name] = style_func

    def __getitem__(self, item):
        """ Given a valid style-ID, retrieve a stored style. If a valid function (callable) is
            supplied, return it in place.

            Parameters
            ----------
            item : Union[Any, Callable[Optional[str], Optional[str]], Optional[str]]
                A valid style-ID or style-function."""
        try:
            return self._store[item]
        except KeyError:
            try:
                _check_style_function(item)
                return item
            except TypeError:
                raise TypeError("Either a valid style name or style-function must be specified")

    def keys(self):
        """  D.keys() -> a set-like object providing a view on D's keys"""
        return self._store.keys()

    def pop(self, *args):
        """ D.pop(k[,d]) -> v, remove specified key and return the corresponding value.
            If key is not found, d is returned if given, otherwise KeyError is raised. """
        if len(args) < 3:
            return self._store.pop(*args)
        else:
            raise TypeError("pop expected at most 2 arguments, got {}".format(len(args)))

    def update(self, *args, **kwargs):
        """ D.update([E, ]**F) -> None.  Update D from dict/iterable E and F.
            If E is present and has a .keys() method, then does:  for k in E: D[k] = E[k]"""
        if len(args) > 1:
            raise TypeError("update expected at most 1 arguments, got %d" % len(args))

        for key, value in dict(*args, **kwargs).items():
            self[key] = value

    def values(self):
        """ D.values() -> an object providing a view on D's values."""
        return self._store.values()

    def items(self):
        """ D.items() -> a set-like object providing a view on D's items"""
        return self._store.items()

store = _Store([(key, getattr(style_store, key)) for key in style_store.__all__])


def add_style(style_name, style_func):
    """ Make available a new function for merging a 'parent' and 'child' docstring.

        Parameters
        ----------
        style_name : Any
            The identifier of the style being logged
        style_func: Callable[[Optional[str], Optional[str]], Optional[str]]
            The style function that merges two docstrings into a single docstring."""
    store[style_name] = style_func


def remove_style(style):
    """ Remove the specified style from the style store.

        Parameters
        ----------
        style: Any
            The inheritance-scheme style ID to be removed."""
    if style in store:
        store.pop(style)


def DocInheritMeta(style="parent", abstract_base_class=False):
    """ A metaclass that merges the respective docstrings of a parent class and of its child, along with their
        properties, methods (including classmethod, staticmethod, decorated methods).

        Parameters
        ----------
        style: Union[Any, Callable[[str, str], str]], optional (default: "parent")
            A valid inheritance-scheme style ID or function that merges two docstrings.

        abstract_base_class: bool, optional(default: False)
            If True, the returned metaclass inherits from abc.ABCMeta.

            Thus a class that derives from DocInheritMeta(style="numpy", abstract_base_class=True)
            will be an abstract base class, whose derived classes will inherit docstrings
            using the numpy-style inheritance scheme.


        Returns
        -------
        custom_inherit.DocInheritorBase"""

    merge_func = store[style]
    metaclass = DocInheritorBase
    metaclass.class_doc_inherit = staticmethod(merge_func)
    metaclass.attr_doc_inherit = staticmethod(merge_func)

    return metaclass if not abstract_base_class else type("abc" + metaclass.__name__, (ABCMeta, metaclass), {})


def doc_inherit(parent, style="parent"):
    """ Returns a function/method decorator that, given `parent`, updates the docstring of the decorated
        function/method based on the specified style and the corresponding attribute of `parent`.

        Parameters
        ----------
        parent : Union[str, Any]
            The docstring, or object of which the docstring is utilized as the
            parent docstring during the docstring merge.

        style : Union[Any, Callable[[str, str], str]], optional (default: "parent")
            A valid inheritance-scheme style ID or function that merges two docstrings.

        Returns
        -------
        custom_inherit.DocInheritDecorator

        Notes
        -----
        `doc_inherit` should always be used as the inner-most decorator when being used in
        conjunction with other decorators, such as `@property`, `@staticmethod`, etc."""

    merge_func = store[style]
    decorator = DocInheritDecorator
    decorator.doc_merger = staticmethod(merge_func)
    return decorator(parent)



from __future__ import absolute_import
from distutils.core import setup
from setuptools import find_packages
import custom_inherit

ver = custom_inherit.__version__
setup(name='custom_inherit',
      version= ver,
      description='A Python package that provides customized docstring inheritance\
       schemes between derived classes and their parents.',
      author='@meowklaski',
      author_email="rsoklaski@gmail.com",
      url='https://github.com/meowklaski/custom_inherit',
      download_url="https://github.com/meowklaski/custom_inherit/tarball/"+ver,
      packages=find_packages(),
      license="MIT"
      )



from __future__ import absolute_import
from abc import abstractproperty
from types import FunctionType

""" Exposes abstract base meta class to be inherited by inheritance-style meta classes.

    This metaclass merges the respective docstrings of a parent class and its child, and their
    properties, methods (including classmethod, staticmethod, decorated methods)

    This merge-style must be implemented via the static methods `class_doc_inherit`
    and `attr_doc_inherit`. See custom_inherit/style_store.py for such implementations."""

__all__ = ["DocInheritorBase"]


class DocInheritorBase(type):
    """ A metaclass that merges the respective docstrings of a parent class and of its child, along with their
        properties, methods (including classmethod, staticmethod, decorated methods).

        This merge-style must be implemented via the static methods `class_doc_inherit`
        and `attr_doc_inherit`, which are set within `custom_inherit.DocInheritMeta`."""
    def __new__(mcs, class_name, class_bases, class_dict):
        # inherit class docstring
        clsdoc = class_dict.get("__doc__", None)
        prnt_cls_doc = None
        for mro_cls in (mro_cls for base in class_bases for mro_cls in base.mro()):
            prnt_cls_doc = mro_cls.__doc__
            if prnt_cls_doc is not None:
                if prnt_cls_doc == "The most base type":
                    prnt_cls_doc = None
                break
        class_dict["__doc__"] = mcs.class_doc_inherit(prnt_cls_doc, clsdoc)

        # inherit docstring for method, static-method, class-method, abstract-method, decorated-method, and property
        for attr, attribute in class_dict.items():
            is_doc_type = isinstance(attribute, (FunctionType, classmethod, staticmethod, property))

            if (attr.startswith("__") and attr.endswith("__")) or not is_doc_type:
                continue

            is_static_or_class = isinstance(attribute, (staticmethod, classmethod))
            child_attr = attribute if not is_static_or_class else attribute.__func__

            prnt_attr_doc = None
            for mro_cls in (mro_cls for base in class_bases
                            for mro_cls in base.mro() if hasattr(mro_cls, attr)):
                prnt_attr_doc = getattr(mro_cls, attr).__doc__

                if prnt_attr_doc is not None:
                    break

            if prnt_attr_doc is None:
                continue

            doc = mcs.attr_doc_inherit(prnt_attr_doc, child_attr.__doc__)
            try:
                child_attr.__doc__ = doc
            except TypeError as err:
                if isinstance(child_attr, property):  # property.__doc__ is read-only in Python 2
                    new_prop = property(fget=child_attr.fget,
                                        fset=child_attr.fset,
                                        fdel=child_attr.fdel,
                                        doc=doc)
                    if isinstance(child_attr, abstractproperty):
                        new_prop = abstractproperty(new_prop)
                    class_dict[attr] = new_prop
                else:
                    raise TypeError(err)

        return type.__new__(mcs, class_name, class_bases, class_dict)

    @staticmethod
    def class_doc_inherit(prnt_cls_doc, child_doc):
        """ Merge the docstrings of a parent class and its child.

            Parameters
            ----------
            prnt_cls_doc: Union[None, str]
            child_doc: Union[None, str]

            Raises
            ------
            NotImplementedError"""
        raise NotImplementedError

    @staticmethod
    def attr_doc_inherit(prnt_attr_doc, child_doc):
        """ Merge the docstrings of method or property from parent class and the corresponding
            attribute of its child.

            Parameters
            ----------
            prnt_cls_doc: Union[None, str]
            child_doc: Union[None, str]

            Raises
            ------
            NotImplementedError

            Notes
            -----
            This works for properties, methods, static methods, class methods, and
            decorated methods/properties."""
        raise NotImplementedError

